/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Düsterloh <mnenhy@tprac.de>.
 *
 * Portions created by the Initial Developer are Copyright (C) 2002-2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Karsten Düsterloh <mnenhy@tprac.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// +-
// |    Written, forged, copied and conceived by Karsten Düsterloh <mnenhy@tprac.de>
// |    Inspired by Moztweak-ROT13 by Marek Wawoczny <Marek.Wawoczny@GmbH.rubikon.pl>
// |         and by Kenny-Speak    by Kohan Ikin
// +-

// |
// |  constants
// |

// global arrays for easy encoding
const gkrsROT13           = { a: "n",  b: "o",  c: "p",  d: "q",  e: "r",  f: "s",  g: "t",  h: "u",
                              i: "v",  j: "w",  k: "x",  l: "y",  m: "z",  n: "a",  o: "b",  p: "c",
                              q: "d",  r: "e",  s: "f",  t: "g",  u: "h",  v: "i",  w: "j",  x: "k",
                              y: "l",  z: "m",
                              A: "N",  B: "O",  C: "P",  D: "Q",  E: "R",  F: "S",  G: "T",  H: "U",
                              I: "V",  J: "W",  K: "X",  L: "Y",  M: "Z",  N: "A",  O: "B",  P: "C",
                              Q: "D",  R: "E",  S: "F",  T: "G",  U: "H",  V: "I",  W: "J",  X: "K",
                              Y: "L",  Z: "M"
                            };
const gkrsKenny           = { // normal to Kenny
                              a:"mmm", b:"mmp", c:"mmf", d:"mpm", e:"mpp", f:"mpf", g:"mfm", h:"mfp",
                              i:"mff", j:"pmm", k:"pmp", l:"pmf", m:"ppm", n:"ppp", o:"ppf", p:"pfm",
                              q:"pfp", r:"pff", s:"fmm", t:"fmp", u:"fmf", v:"fpm", w:"fpp", x:"fpf",
                              y:"ffm", z:"ffp",
                              "":"mmmmpp",     "":"ppfmpp",     "":"fmfmpp",     "":"fmmffp",
                              A:"Mmm", B:"Mmp", C:"Mmf", D:"Mpm", E:"Mpp", F:"Mpf", G:"Mfm", H:"Mfp",
                              I:"Mff", J:"Pmm", K:"Pmp", L:"Pmf", M:"Ppm", N:"Ppp", O:"Ppf", P:"Pfm",
                              Q:"Pfp", R:"Pff", S:"Fmm", T:"Fmp", U:"Fmf", V:"Fpm", W:"Fpp", X:"Fpf",
                              Y:"Ffm", Z:"Ffp",
                              "":"Mmmmpp",     "":"Ppfmpp",     "":"Fmfmpp",
                              // Kenny to normal
                              mmm:"a", mmp:"b", mmf:"c", mpm:"d", mpp:"e", mpf:"f", mfm:"g", mfp:"h",
                              mff:"i", pmm:"j", pmp:"k", pmf:"l", ppm:"m", ppp:"n", ppf:"o", pfm:"p",
                              pfp:"q", pff:"r", fmm:"s", fmp:"t", fmf:"u", fpm:"v", fpp:"w", fpf:"x",
                              ffm:"y", ffp:"z",
                              Mmm:"A", Mmp:"B", Mmf:"C", Mpm:"D", Mpp:"E", Mpf:"F", Mfm:"G", Mfp:"H",
                              Mff:"I", Pmm:"J", Pmp:"K", Pmf:"L", Ppm:"M", Ppp:"N", Ppf:"O", Pfm:"P",
                              Pfp:"Q", Pff:"R", Fmm:"S", Fmp:"T", Fmf:"U", Fpm:"V", Fpp:"W", Fpf:"X",
                              Ffm:"Y", Ffp:"Z"
                            };
const gkrsMorse           = { // normal to Morse
                              "a":".-",       "b":"-...",     "c":"-.-.",     "d":"-..",
                              "e":".",        "f":"..-.",     "g":"--.",      "h":"....",
                              "i":"..",       "j":".---",     "k":"-.-",      "l":".-..",
                              "m":"--",       "n":"-.",       "o":"---",      "p":".--.",
                              "q":"--.-",     "r":".-.",      "s":"...",      "t":"-",
                              "u":"..-",      "v":"...-",     "w":".--",      "x":"-..-",
                              "y":"-.--",     "z":"--..",
                              "1":".----",    "2":"..---",    "3":"...--",    "4":"....-",
                              "5":".....",    "6":"-....",    "7":"--...",    "8":"---..",
                              "9":"----.",    "0":"-----",
                              ",":"--..--",   ":":"---...",   "?":"..--..",   "'":".----.",
                              '"':".-..-.",   "/":"-..-.",    "+":".-.-.",    "=":"-...-",
                              "(":"-.--.-",   "@":".--.-.",
                              // Morse to normal
                              ".-"    :"a",   "-..."  :"b",   "-.-."  :"c",   "-.."   :"d",
                              "."     :"e",   "..-."  :"f",   "--."   :"g",   "...."  :"h",
                              ".."    :"i",   ".---"  :"j",   "-.-"   :"k",   ".-.."  :"l",
                              "--"    :"m",   "-."    :"n",   "---"   :"o",   ".--."  :"p",
                              "--.-"  :"q",   ".-."   :"r",   "..."   :"s",   "-"     :"t",
                              "..-"   :"u",   "...-"  :"v",   ".--"   :"w",   "-..-"  :"x",
                              "-.--"  :"y",   "--.."  :"z",
                              ".----" :"1",   "..---" :"2",   "...--" :"3",   "....-" :"4",
                              "....." :"5",   "-...." :"6",   "--..." :"7",   "---.." :"8",
                              "----." :"9",   "-----" :"0",
                              ".-.-.-":".",   "--..--":",",   "---...":":",   "..--..":"?",
                              ".----.":"'",   ".-..-.":'"',   "-....-":"-",   "-..-." :"/",
                              ".-.-." :"+",   "-...-" :"=",   "-.--.-":"(",   ".--.-.":"@",
                              // special chars
                              "dot"  : ".-.-.-",
                              "minus": "-....-",
                              "\n"   : "\n",
                              "\r"   : "\r",
                              "\n\r" : "\n\r",
                              "\t"   : "\t",
                              "\v"   : "\v",
                              "\f"   : "\f",
                              " "    : "",
                              // unreversable characters
                              ")":"-.--.-",     // coded as "("
                              "":".- .",       // coded as "ae"
                              "":"--- .",      // coded as "oe"
                              "":"..- .",      // coded as "ue"
                              "":"... --.."    // coded as "sz"
                            };
const gkrs1337            = { // normal to 1337
                              "a":"4",    "b":"|>",   "c":"(",      "d":"|)",   "e":"3",
                              "f":"|=",   "g":"6",    "h":"|-|",    "i":"!",    "j":",|",
                              "k":"|<",   "l":"1",    "m":"|\\/|",  "n":"/\\/", "o":"0",
                              "p":"|*",   "q":"9",    "r":"|2",     "s":"5",    "t":"7",
                              "u":"|_|",  "v":"\\/",  "w":"\'//",   "x":"><",   "y":"`/",   "z":"2",
                              // 1337 to normal
                              "4":"a",    "/\\":"a",  "@":"a",
                              "|3":"b",   "8":"b",    "|>":"b",
                              "[":"c",    "":"c",    "(":"c",
                              "|)":"d",   "o|":"d",
                              "3":"e",    "&":"e",
                              "|=":"f",
                              "6":"g",
                              "#":"h",    "|-|":"h",  "]~[":"h",
                              "!":"i",    "|":"i",
                              ",|":"j",   "_|":"j",
                              "|<":"k",   "|{":"k",
                              "1":"l",    "|_":"l",
                              "//.":"m",  "^^":"m",   "|v|":"m",    "|\\/|":"m",   "(u)":"m",
                              "/\\/\\":"m",
                              "//":"n",   "^/":"n",   "|\\|":"n",   "/\\/":"n",
                              "0":"o",    "()":"o",
                              "|*":"p",   "|o":"p",   "|\"":"p",
                              "9":"q",    "(,)":"q",
                              "|2":"r",
                              "5":"s",    "$":"s",
                              "7":"t",    "+":"t",    "-|-":"t",
                              "(_)":"u",  "|_|":"u",
                              "\\/":"v",  "<":"v",
                              "\\^/":"w", "(n)":"w",  "\\/\\/":"w", "'//":"w",
                              "><":"x",   "}{":"x",
                              "'/":"y",   "`/":"y",   "\\-/":"y",
                              "2":"z",
                              // german characters (unreversable)
                              "":"43",     // coded as "ae"
                              "":"03",     // coded as "oe"
                              "":"|_|3",   // coded as "ue"
                              "":"52"      // coded as "sz"
                            };
const gkx1337Char         = /^[a-z]$/;
const gkxEncURIChar       = /^%[a-f0-9]{2}$/i;
const gksEncUnicodeBase   = "000000000000000000000000000000000000"; // 36 * 0


// |
// |  variables
// |
var goDocument = null;  // the current document
var goWindow   = null;  // the current document

// Init Prompt texts
var gsEvalTitle     = null;
var gsEvalPrompt    = null;
var gsUnicodeTitle  = null;
var gsUnicodePrompt = null;

// CodecEval will set the codec result into the first non-whitespace text node
// and set the rest to ""
//  goEvalResult == null:                            no CodecEval
//               == {bNew: <bool>, sText: <string>}: have CodecEval, only use it when bNew == true
var goEvalResult = null;

// custom Unicode conversion needs to know base and minimum length
var miCodecUnicodeBase = 0;


// |
// |  Init
// |
function Init(aoMnenhy)
{
  // Init Prompt texts
  const koPropertiesCodecs = aoMnenhy.InitStringBundle("chrome://mnenhy/locale/mnenhy-codecs.properties");
  gsEvalTitle     = aoMnenhy.GetString("EvalTitle",     koPropertiesCodecs);
  gsEvalPrompt    = aoMnenhy.GetString("EvalPrompt",    koPropertiesCodecs);
  gsUnicodeTitle  = aoMnenhy.GetString("UnicodeTitle",  koPropertiesCodecs);
  gsUnicodePrompt = aoMnenhy.GetString("UnicodePrompt", koPropertiesCodecs);

  // Initialize gaps in mkrs1337
  // so that we can find the longest match easily
  for (var s in gkrs1337)
  {
    const kis = s.length;
    for (var i = 1; i < kis; ++i)
    {
      const ksf = s.substr(0, i);
      if (!(ksf in gkrs1337))
        gkrs1337[ksf] = null;
    }
  }
}


// |
// |  functions
// |

// when the codecs are called, we may have lost the original global
// Javascript objects, so we need to use those of the calling window
function Codec(aoCaller, aoEvent, asCodec, aiSubmenuDepth)
{ // get current document and window
  goDocument = aoCaller.ownerDocument;
  goWindow   = goDocument.defaultView;
  goMnenhy.SetWaitCursor(goWindow);

  // get the parent id for the (sub)menu entry
  if (!aiSubmenuDepth)
    aiSubmenuDepth = 3;
  else
    aiSubmenuDepth += 3;
  var oNode = aoCaller;
  for (var i = 0; (i < aiSubmenuDepth) && oNode; ++i)
    oNode = oNode.parentNode;
  const ksID = oNode ? oNode.id : aoCaller; /* if all fails, fail safely with aoCaller */

  // kill the calling popup
  // actually, I wonder why it is shown here still at all :| - it wasn't in the past
  const koPopUp = goDocument.getElementById(ksID);
  var bState = null;
  if (koPopUp)
  {
    bState = koPopUp.collapsed;
    koPopUp.collapsed = true;
  }

  // differentiate between editor and messagepane
  try
  {
    switch (ksID)
    { // Mozilla places
      case "messagePaneContext":      CallFromBrowser(asCodec, "messagepane"  ); break; // message display
      case "viewSourceContextMenu":   CallFromBrowser(asCodec, "content"      ); break; // source view
      case "contentAreaContextMenu":  CallFromBrowser(asCodec, "content"      ); break; // browser
      case "msgComposeContext":       CallFromEditor (asCodec, "content-frame"); break; // message editor
      case "MnenhyMailEditor":                                                          // message editor
      case "MnenhyChatzillaSingle":   
      case "MnenhyChatzillaMulti":    CallFromTextbox(asCodec                 ); break; // Chatzilla
      case "context:messages":        CallFromBrowser(asCodec, "content"      ); break; // Chatzilla
      // Mnenhy places
      case "MnenhyTextPopup":         CallFromIframe (asCodec                 ); break;
    }
  }
  catch(oEx)
  {
    goMnenhy.Log("Codec: exception="+oEx)
  }

  // reactivate popup - otherwise we won't see another one ;-)
  if (koPopUp)
    koPopUp.collapsed = bState;

  goMnenhy.UnsetWaitCursor(goWindow);
  goDocument = null;
  aoEvent.stopPropagation(); // do not bubble
  return true;
}


function CallFromEditor(asCodec, asEditorID)
{
  const koContentFrame = goDocument.getElementById(asEditorID); // <editor>
  if (koContentFrame)
  {
    const koEditor     = koContentFrame.getEditor(koContentFrame.contentWindow);
    if (koEditor && (koEditor instanceof Components.interfaces.nsIPlaintextEditor))
    {
      koEditor.beginTransaction();
      ProcessCodec(asCodec,
                   koEditor.rootElement,
                   koEditor.selection,
                   koEditor);
      koEditor.endTransaction();
    }
  }
}


function CallFromBrowser(asCodec, asBrowserID)
{ // get the owning document of the click (might be in a frame!)
  const koContentFrame = goDocument.popupNode.ownerDocument;
  if (koContentFrame)
  {
    ProcessCodec(asCodec,
                 koContentFrame.getElementsByTagName("body")[0],
                 koContentFrame.defaultView.getSelection());
  }
}


function CallFromTextbox(asCodec)
{ // the popupNode is either a textarea or an input field,
  // but the view's selection is usually not defined
  var oBase = goDocument.popupNode;
  if (GetBaseName(oBase) == "input")
    oBase = oBase.parentNode;
  const koRange = goDocument.createRange();
  koRange.setStart(oBase, 0); // INPUT is first child node, TEXTAREA does not care
  koRange.setEnd  (oBase, 0); // must be equal to start 

  const koSelection =
  {
    rangeCount:   1,
    isCollapsed:  true,
    getRangeAt:   function(ignored){return koRange;},
    collapse:     function(ignored, ignored2){},
    extend:       function(ignored, ignored2){},
    selectAllChildren: function(ignored){},
    removeAllRanges:   function(){},
    toString:     function(){return "";}
  };
  ProcessCodec(asCodec,
               goDocument.popupNode,
               koSelection);
}


function CallFromIframe(asCodec)
{ // we're called from a Mnenhy header, and that's an iframe - but which one?
  const ksObserverID = goDocument.popupNode.ownerDocument.documentElement.getAttribute("mnenhyObserverID");
  const koHeader     = goMnenhy.GetObserver(ksObserverID);
  const koIframe     = goDocument.getAnonymousElementByAttribute(koHeader, "anonid", "HTMLValue");
  if (koIframe)
  {
    ProcessCodec(asCodec,
                 koIframe.contentWindow.document.getElementsByTagName("body")[0],
                 koIframe.contentWindow.getSelection());
  }
}


// apply codec to aoTextNode
// If aoEditor is specified, the aoSelection is set to (aoTextNode, aiStartOffset, aiEndOffset)
//   and changed by insertText to allow for undo. This will destroy the current selection!
// Otherwise, the node text is changed directly via replaceData.
// If aoTextNode is an <input> or <textarea> element, the new selection will be set here also.
// (We assume that no call with aoEditor will use <input> or <textarea>!)
// Returns new aiEndOffset or -1 if unchanged.
function ApplyCodec(aoTextNode,
                    asCodec,
                    aoSelection,
                    aiStartOffset /* = null */,
                    aiEndOffset   /* = null */,
                    aoEditor      /* = null */)
{
//dump("ApplyCodec("+aoTextNode.nodeName+")["+aiStartOffset+","+aiEndOffset+"]"+aoTextNode.nodeValue+"\n");
  if (aiStartOffset !== null && aiEndOffset !== null && aiStartOffset == aiEndOffset)
  { // canonify empty range
    aiStartOffset = aiEndOffset = null;
  }  

  const kbIsText = aoTextNode.nodeType == Node.TEXT_NODE;
  var sSource, bHaveInputSelection = !kbIsText;
  if (kbIsText)
  {
    sSource = aoTextNode.nodeValue;
    if (aiStartOffset == null)
      aiStartOffset = 0;
    if (aiEndOffset == null)
      aiEndOffset = sSource.length;
  }
  else
  {
    sSource = aoTextNode.value;
    aiStartOffset = (aiStartOffset !== null) ? aoTextNode.selectionStart : 0;
    aiEndOffset   = (aiEndOffset   !== null) ? aoTextNode.selectionEnd   : aoTextNode.textLength;
  }

  var iEndOffsetRev = aiEndOffset;
  if (aiEndOffset > aiStartOffset)
  {
    var iOldLength = aiEndOffset - aiStartOffset;
    var sOldText   = sSource.substring(aiStartOffset, aiEndOffset);
    var sNewText   = "";

    if (goEvalResult)
    { // use read value, but only once)
      if (goEvalResult.bNew && sOldText.search(/[^\s]/) >= 0 )
      {
        sNewText = goEvalResult.sText;
        goEvalResult.bNew = false;
      }
    }
    else
    {
      sNewText = this[asCodec](sOldText);
    }
//goMnenhy.Log("["+aiStartOffset+","+aiEndOffset+"]"+sSource+"\n==>"+sNewText);
    if (aoEditor)
    { // set the selection to just the node
      aoSelection.collapse(aoTextNode, aiStartOffset);
      aoSelection.extend  (aoTextNode, aiEndOffset  );
      if (sNewText)
        aoEditor.insertText(sNewText);
      else
        aoEditor.deleteSelection(aoEditor.eNext);
    }
    else
    { // just change the text - no editor, no undo
      if (kbIsText)
        aoTextNode.replaceData(aiStartOffset, aiEndOffset - aiStartOffset, sNewText);
      else
        aoTextNode.value = sSource.substr(0, aiStartOffset) + sNewText + sSource.substr(aiEndOffset);
    }

    // set new aiEndOffset
    if (sNewText.length != iOldLength)
      iEndOffsetRev += sNewText.length - iOldLength;
    if (!kbIsText /* bHaveInputSelection */)
    {
      aoTextNode.selectionStart = aiStartOffset;
      aoTextNode.selectionEnd   = iEndOffsetRev;
    }
  }
  return iEndOffsetRev;
}


// return the index in the parent's childNodes
// (we expect its existence!)
function FindIndexInParent(aoNode)
{
  const koh = aoNode.parentNode.childNodes;
  for (var ih = 0; ih < koh.length; ++ih)
  {
    if (koh[ih] == aoNode)
      return ih;
  }
  return -1;  // error
}


// get the lowercase name of the node
function GetBaseName(aoNode)
{
  var sName = "";
  if (aoNode)
  {
    sName = aoNode.localName || aoNode.nodeName;
    if (sName)
      sName = sName.toLowerCase();
  }
  return sName;
}


// nsIDOMNodeFilter function for ProcessCodec
// allow <#text>, <textarea> and <input> elements
function ShowTextNodes(aoNode)
{
//dump("ShowTextNodes: ");
//dump("localName="+aoNode.localName+", aoNode.nodeName="+aoNode.nodeName+"\n");
  const ksName = GetBaseName(aoNode);
//dump("("+ksName+")=");
  switch (ksName)
  {
    case "#text": // ignore #text childs of <textarea>
      if (GetBaseName(aoNode.parentNode) == "textarea")
      {
//dump("==> rejected!\n")
        return NodeFilter.FILTER_REJECT;
      }
    case "textarea":
//dump("==> accepted!\n")
      return NodeFilter.FILTER_ACCEPT;

    case "input":
      if (aoNode.getAttribute("type") && aoNode.getAttribute("type").toLowerCase() != "text")
      {
//dump("==> rejected!\n")
        return NodeFilter.FILTER_REJECT;
      }
//dump("==> accepted!\n")
      return NodeFilter.FILTER_ACCEPT;
  }
//dump("==> skipped!\n")
  return NodeFilter.FILTER_SKIP;
}


function ProcessCodec(asCodec, aoRootElement, aoSelection, aoEditor /* = null */)
{ // it isn't possible to have multiple ranges in the selections here
  var oTreeWalker = null;
  var oNextNode   = null;
  var oThisNode   = null;
  goEvalResult = null;

  var bProcessAll = aoSelection.rangeCount != 1;
  if (!bProcessAll) // just one range
  {
    var oRange          = aoSelection.getRangeAt(0);
    var oStartContainer = oRange.startContainer;
    var iStartOffset    = oRange.startOffset;
    var iEndOffset      = oRange.endOffset;

    bProcessAll = aoSelection.isCollapsed;
    if (bProcessAll)
    { // no real selection, but maybe a <teaxtarea> or <input> field
      // with a selection inside...
      bProcessAll = GetBaseName(oStartContainer) != "textarea";
      if (bProcessAll)
      {
        if (iStartOffset < oStartContainer.childNodes.length)
          oStartContainer = oStartContainer.childNodes[iStartOffset];
        bProcessAll = GetBaseName(oStartContainer) != "input";
      }

      if (!bProcessAll) // just one textarea or input field
      {
        iStartOffset = oStartContainer.selectionStart;
        iEndOffset   = oStartContainer.selectionEnd;
        // special handling for some codecs
        switch (asCodec)
        {
          case "CodecEval":
            if (!PromptForEval(oStartContainer.value.substring(iStartOffset, iEndOffset)))
              return; // cancel codec
            break;
          case "EncUnicode2Custom":
          case "DecUnicode2Custom":
            if (!PromptForUnicodeBase())
              return; // cancel codec
            break;
        }
        ApplyCodec(oStartContainer, asCodec, aoSelection, iStartOffset, iEndOffset, aoEditor);
      }
    }
    else // have real selection
    { // special handling for some codecs
      switch (asCodec)
      {
        case "CodecEval":
          if (!PromptForEval(oRange.toString()))  // this will not take care of contained
            return; // cancel codec               // textareas or input fields!
          break;
        case "EncUnicode2Custom":
        case "DecUnicode2Custom":
          if (!PromptForUnicodeBase())
            return; // cancel codec
          break;
      }
      // remember the selection boundaries
      var oStartParent  = oStartContainer.parentNode;
      var iStartIndex   = FindIndexInParent(oStartContainer);

      var oEndContainer = oRange.endContainer;
      var oEndParent    = oEndContainer.parentNode;
      var iEndOffsetRev = iEndOffset;
      var iEndIndex     = FindIndexInParent(oEndContainer);

      if ((oStartContainer == oEndContainer) && (oStartContainer.nodeType == Node.TEXT_NODE))
      { // just one text node
        iEndOffsetRev = ApplyCodec(oStartContainer, asCodec, aoSelection, iStartOffset, iEndOffset, aoEditor);
      }
      else
      { // different boundary containers or not text nodes

        // process the boundaries first, because the oTreeWalker may destroy them
        if ((oStartContainer.nodeType == Node.TEXT_NODE) && (iStartOffset > 0))
        { // start lies within TEXT_NODE
          ApplyCodec(oStartContainer, asCodec, aoSelection, iStartOffset, null, aoEditor);
        }
        if (oEndContainer.nodeType == Node.TEXT_NODE)
        { // set the iEndOffsetRev
          iEndOffsetRev = -1; // recalculate after oTreeWalker
          if (iEndOffset < oEndContainer.nodeValue.length)
          { // end lies within TEXT_NODE
            iEndOffsetRev = ApplyCodec(oEndContainer, asCodec, aoSelection, null, iEndOffset, aoEditor);
          }
        }

        // now process the in-between nodes
        oTreeWalker = goDocument.createTreeWalker(oRange.commonAncestorContainer,
                                                  NodeFilter.SHOW_ELEMENT | NodeFilter.SHOW_TEXT,
                                                  ShowTextNodes,
                                                  true);
        if (oTreeWalker)
        { // process all text nodes after the start node and before the final node
          for (oNextNode = oTreeWalker.nextNode(); oNextNode; )
          { // ApplyCodec may destroy oNextNode, so get the next node before changing
            oThisNode = oNextNode;
            oNextNode = oTreeWalker.nextNode();

            // select whole text to exclude boundaries
            var oCursorRange = goDocument.createRange();
            oCursorRange.setStart(oThisNode, 0);
            oCursorRange.setEnd  (oThisNode, (oThisNode.nodeType == Node.TEXT_NODE)
                                             ? oThisNode.nodeValue.length : 0);
            if (    (oCursorRange.compareBoundaryPoints(Range.START_TO_START, oRange) >= 0)
                 && (oCursorRange.compareBoundaryPoints(Range.END_TO_END,     oRange) <= 0))
            {
              ApplyCodec(oThisNode, asCodec, aoSelection, null, null, aoEditor);
            }
            oCursorRange.detach();
          } // for oNextNode
        } // if oTreeWalker

        // recalculate iEndOffsetRev
        if (iEndOffsetRev < 0)
          iEndOffsetRev = oEndParent.childNodes[iEndIndex].nodeValue.length;
      } // different boundary containers or not text nodes

      // restore selection (need to recompute end offset in text nodes)
      // only extend selection if there is something to mark (else .extend will bail)
      aoSelection.collapse(oStartParent.childNodes[iStartIndex], iStartOffset);
      if (oStartParent != oEndParent || iStartIndex != iEndIndex || iStartOffset != iEndOffsetRev)
        aoSelection.extend(oEndParent.childNodes[iEndIndex], iEndOffsetRev);
    } // real selection
  }

  if (bProcessAll)
  { // no selection, so do it for all text children of the aoRootElement

    // special handling for some codecs
    switch (asCodec)
    {
      case "CodecEval":
        {
          aoSelection.selectAllChildren(aoRootElement);
          var sEval = aoSelection.toString();
          aoSelection.removeAllRanges();
          aoSelection.collapse(aoRootElement, 0); // put cursor at begin of document
          if (!PromptForEval(sEval))
            return; // cancel codec
        }
        break;
      case "EncUnicode2Custom":
      case "DecUnicode2Custom":
        if (!PromptForUnicodeBase())
          return; // cancel codec
        break;
    }

    oTreeWalker = goDocument.createTreeWalker(aoRootElement,
                                              NodeFilter.SHOW_ELEMENT | NodeFilter.SHOW_TEXT,
                                              ShowTextNodes,
                                              true);
    if (oTreeWalker)
    {
      for (oNextNode = oTreeWalker.nextNode(); oNextNode; )
      {
        oThisNode = oNextNode;
        oNextNode = oTreeWalker.nextNode();
        ApplyCodec(oThisNode, asCodec, aoSelection, null, null, aoEditor);
      }
    } // have treewalker
  } // no selection
}


//
// codecs
//
// some codecs use arrays for performance reasons


// ROT-13
// unknown combinations are ignored
function CodecROT13(asContent)
{
  var rs = [];
  for (var i = 0; i < asContent.length; ++i)
  {
    var c = asContent[i];
    rs.push((c in gkrsROT13) ? gkrsROT13[c] : c);
  }
  return rs.join("");
}


// Base64
// if encoding/decoding fails, return old content
function EncBase64(asContent)
{
  try
  {
    return btoa(asContent);
  }
  catch(e)
  {
    goMnenhy.Log("EncBase64: Error:\n"+e);
    return asContent;
  }
}
function DecBase64(asContent)
{ // ignore any whitespace
  asContent = asContent.replace(/\s+/g, "");
  // only A-Za-z0-9+/= are allowed
  const kxAllowed    =  /[A-Za-z0-9\+\/=]/;
  const kxNotAllowed = /[^A-Za-z0-9\+\/=]/;
  var rs = [];
  for (var i = 0; i < asContent.length;)
  { // search for the next not allowed char
    var p = asContent.substr(i).search(kxNotAllowed);
    if (p < 0)
      p = asContent.length;
    // translate allowed part
    var s = asContent.substring(i, i + p);
    try
    {
      rs.push(atob(s));
    }
    catch(e)
    {
      goMnenhy.Log("DecBase64: Error:\n"+e);
      rs.push(s);
    }
    i += p;

    // process unallowed chars
    // search for the next allowed char
    p = asContent.substr(i).search(kxAllowed);
    if (p < 0)
      p = asContent.length;
    // copy not allowed part
    rs.push(asContent.substring(i, i + p));
    i += p;
  } // for all asContent chars
  return rs.join("");
}


// helper function for 'escape/encodeURI all'
// all characters in asContent are replaced by %charcode to allow for
// decoding by unescape/decodeURI(Component)
// already existing encodings will NOT be touched
function EncURIHelper(asContent)
{
  var rs = [], sEnc;
  const kiLen = asContent.length;
  for (var i = 0; i < kiLen; ++i)
  {
    sEnc = asContent.substr(i,3); // potentially already encoded
    if (sEnc.search(gkxEncURIChar))
    { // encode normal Unicode character
      sEnc = "%" + ("00" + sEnc.charCodeAt(0).toString(16)).substr(-2);
    }
    else
      i += 2;
    rs.push(sEnc)
  }
  return rs.join("");
}

// escape (all) / unescape
function EncEscape(asContent)
{
  return escape(asContent);
}
function EncEscapeAll(asContent)
{
  return EncURIHelper(escape(asContent));
}
function DecUnescape(asContent)
{
  return unescape(asContent);
}


// encodeURI / decodeURI
function EncURI(asContent)
{
  return encodeURI(asContent);
}
function DecURI(asContent)
{
  return decodeURI(asContent);
}


// encodeURIComponent (all) / decodeURIComponent
function EncURIComponent(asContent)
{
  return encodeURIComponent(asContent);
}
function EncURIAll(asContent)
{
  return EncURIHelper(encodeURIComponent(asContent));
}
function DecURIComponent(asContent)
{
  return decodeURIComponent(asContent);
}


// uuencode / uudecode
function EncUUEncode(asContent)
{
  var rs = [];
  for (var i = 0; i < asContent.length; i += 3)
  {
    var iA = asContent.charCodeAt(i);
    var iB = asContent.charCodeAt(i+1);
    if (isNaN(iB))
      iB = 0;
    var iC = asContent.charCodeAt(i+2);
    if (isNaN(iC))
      iC = 0;
    rs.push(String.fromCharCode(0x20 + (( iA >> 2                     ) & 0x3F),
                                0x20 + (((iA << 4) | ((iB >> 4) & 0xF)) & 0x3F),
                                0x20 + (((iB << 2) | ((iC >> 6) & 0x3)) & 0x3F),
                                0x20 + (( iC                          ) & 0x3F)));
  }
  return rs.join("");
}
function UUDecodeBuffer(aBuffer, aTarget)
{ // helper function for DecUUDecode;
  // decoded 0x00-Bytes will be *ignored*,
  // so this implementation does *not* suffice for real binary data
  var iA = ((aBuffer[1] >> 4) | (aBuffer[0] << 2)) & 0xFF;
  var iB = ((aBuffer[2] >> 2) | (aBuffer[1] << 4)) & 0xFF;
  var iC = ((aBuffer[3]     ) | (aBuffer[2] << 6)) & 0xFF;
  if (iA)
    aTarget.push(String.fromCharCode(iA));
  if (iB)
    aTarget.push(String.fromCharCode(iB));
  if (iC)
    aTarget.push(String.fromCharCode(iC));
}
function DecUUDecode(asContent)
{ // decoded 0x00-Bytes will be *ignored*,
  // so this implementation does *not* suffice for real binary data
  // other unknown bytes are tunneled through
  var rs  = [];
  var rsb = []; // decoding buffer
  var isp = 0;  // next valid element in buffer

  for (var i = 0; i < asContent.length; ++i)
  {
    var is  = asContent.charCodeAt(i) - 0x20;
    var bnv = ((is < 0x00) || (is > 0x3F));
    if (bnv)
    { // invalid char, enforce decoding now
      for (;isp < 4; ++isp)
        rsb[isp] = 0;
    }
    else
    { // store valid char
      rsb[isp++] = is;
    }
    if (isp == 4)
    { // enough data or invalid char
      UUDecodeBuffer(rsb, rs);
      isp = 0;
      if (bnv)
        rs.push(asContent[i]);
    }
  }
  if (isp)
  {
    for (;isp < 4; ++isp)
      rsb[isp] = 0;
    UUDecodeBuffer(rsb, rs);
  }
  return rs.join("");
}


// simply reverse the string
function CodecReverse(asContent)
{
  var rs = [];
  for (var i = asContent.length - 1; i >= 0; --i)
    rs.push(asContent[i]);
  return rs.join("");
}


// Kenny-speak
// unknown combinations are tunneled through
function EncKenny(asContent)
{
  var rs = [];
  var s  = "";
  for (var j = 0; j < asContent.length; ++j)
  {
    s = asContent[j];
    rs.push((s in gkrsKenny) ? gkrsKenny[s] : s);
  }
  return rs.join("");
}
function DecKenny(asContent)
{
  var rs = [];
  var s  = "";
  for (var j = 0; j < asContent.length; j += s.length)
  {
    s = asContent.substr(j, 3);    // Kenny generated three chars
    rs.push((s in gkrsKenny) ? gkrsKenny[s] : (s = s[0]));
  }
  return rs.join("");
}


// Morse code
// unknown characters are dropped
// unknown combinations and whitespaces are spaced
// case is lost
function EncMorse(asContent)
{
  var rs = [];
  asContent = asContent.toLowerCase();
  for (var j = 0; j < asContent.length; ++j)
  {
    var s = asContent[j];
    switch (s)
    { // handle special chars
      case ".": s = "dot";   break;
      case "-": s = "minus"; break;
    }
    // unknown characters are dropped
    if (s in gkrsMorse)
      rs.push(gkrsMorse[s] + " ");
  }
  return rs.join("");
}
function DecMorse(asContent)
{ // only dot and hyphen are valid Morse characters
  // the first space after a valid Morse character is a delimiter and dropped hence
  const kxAllowed    =  /[.-]/;
  const kxNotAllowed = /[^.-]/;
  var rs = [];
  for (var i = 0; i < asContent.length;)
  { // search for the next not allowed char
    var p = asContent.substr(i).search(kxNotAllowed);
    if (p < 0)
      p = asContent.length;
    // translate allowed part; unknown combinations are set to spaces
    var s = asContent.substring(i, i + p);
    if (s)
      rs.push((s in gkrsMorse) ? gkrsMorse[s] : " ");
    i += p;

    // process unallowed chars
    // search for the next allowed char
    p = asContent.substr(i).search(kxAllowed);
    if (p < 0)
      p = asContent.length;
    // copy not allowed part, but remove the first (non breaking) space
    // (the messagepane will show the first space rs a non-breaking space,
    //  if format=flowed display is active)
    s = asContent.substring(i, i + p).replace(/^(\u00A0?) (.*)/g, "$1$2");
    if (s)
      rs.push(s);
    i += p;
  } // for all asContent chars
  return rs.join("");
}


// Upper Case
// convert to upper case
function EncUpperCase(asContent)
{
  return asContent.toUpperCase();
}


// Lower Case
// convert to lower case
function EncLowerCase(asContent)
{
  return asContent.toLowerCase();
}


// Evaluate the user-defined expression
function PromptForEval(asContent)
{
  var sEval = null;
  try
  { // PromptService will return null if the user pressed Cancel
    sEval = goMnenhy.PromptService(gsEvalPrompt, "eval(asContent)", gsEvalTitle, goWindow);
    goEvalResult = {bNew: true, sText: sEval ? String(eval(sEval)) : ""};
  }
  catch(err)
  { // dump error
    const ks  = "Error when applying Mnenhy::CodecEval (invalid Javascript?):\n";
    const ksx = "\nExpression: "
    goMnenhy.ConsoleService(ks + err + ksx + sEval);
    goMnenhy.Log(",-- " + ks + err + ksx + sEval + "\n`--");
    sEval = null;  // cancel this action
  }
  if (sEval === null)
  { // error or user cancel
    goEvalResult = null;
    return false;
  }
  return true;
}


// 1337-5|*34|<
// - case-insensitive creation
// - ambiguous, so we prefer longest match - and may fail :|
// - unmatched codes remain undecoded
function Enc1337(asContent)
{
  asContent = asContent.toLowerCase();
  const kiMax = asContent.length;
  var rs = [];

  for (var i = 0; i < kiMax; ++i)
  { // only process letters
    const ksc = asContent[i];
    if (!ksc.search(gkx1337Char) && (ksc in gkrs1337))
      rs.push(gkrs1337[ksc]);
    else
      rs.push(asContent[i])
  }
  return rs.join("");
}
function Dec1337(asContent)
{
  const kiMax = asContent.length;
  var rs = [];

  for (var i = 0; i < kiMax; ++i)
  { // find the longest match
    const kieMax = kiMax - i;
    var sFound = "";
    for (var ie = 1, s = "";
         (ie <= kieMax) && ((s = asContent.substr(i, ie)) in gkrs1337);
         ++ie, sFound = s);
    var sNew = "";
    if (sFound && sFound.search(gkx1337Char))
      sNew = gkrs1337[sFound];

    if (sNew)
    {
      rs.push(sNew);
      i += (sFound.length - 1);
    }
    else
    {
      rs.push(asContent[i])
    }
  }
  return rs.join("");
}



//
// Unicode encoders
//

// parseInt and toString allow bases between 2 and 36
function PromptForUnicodeBase()
{
  var iBase = 0;
  while ( (iBase !== null) && ((iBase < 2) || (iBase > 36)) )
    iBase = goMnenhy.PromptService(gsUnicodePrompt, "10", gsUnicodeTitle, goWindow);
  if (iBase)
  {
    giCodecUnicodeBase = iBase;
    return true;
  }
  return false
}

// convert the bytes of asContent into space separated numbers
// to the given base and minimum length
function EncUnicode2Base(asContent, aiBase, aiMinLength)
{
  const ksStuff = gksEncUnicodeBase.substr(-aiMinLength);
  const kiMax = asContent.length;
  var rs = [];
  for (var i = 0; i < kiMax; ++i)
  {
    var sCode = String(asContent.charCodeAt(i).toString(aiBase));
    if (sCode.length < aiMinLength)
      sCode = (ksStuff + sCode).substr(-aiMinLength);
    rs.push(sCode);
  }
  return rs.join(" ");
}

function EncUnicode2Bin(asContent)
{
  return EncUnicode2Base(asContent,  2, 8);
}
function EncUnicode2Oct(asContent)
{
  return EncUnicode2Base(asContent,  8, 1);
}
function EncUnicode2Dec(asContent)
{
  return EncUnicode2Base(asContent, 10, 1);
}
function EncUnicode2Hex(asContent)
{
  return EncUnicode2Base(asContent, 16, 2);
}
function EncUnicode2Custom(asContent)
{
  return EncUnicode2Base(asContent, giCodecUnicodeBase, 1);
}

// Unicode decoders
// interpret space separated numbers in asContent as character codes
function DecBase2Unicode(asContent, aiBase)
{
  var rs = [];
  const krsn  = asContent.split(/\s+/);
  const kiMax = krsn.length;
  for (var i = 0; i < kiMax; ++i)
  {
    var sDec = krsn[i];
    const ki = parseInt(sDec, aiBase);
    if (!isNaN(ki))
    { // convert it
      sDec = String.fromCharCode(ki);
    }
    rs.push(sDec);
  }
  return rs.join("");
}

function DecBin2Unicode(asContent)
{
  return DecBase2Unicode(asContent,  2);
}
function DecOct2Unicode(asContent)
{
  return DecBase2Unicode(asContent,  8);
}
function DecDec2Unicode(asContent)
{
  return DecBase2Unicode(asContent, 10);
}
function DecHex2Unicode(asContent)
{
  return DecBase2Unicode(asContent, 16);
}
function DecUnicode2Custom(asContent)
{
  return DecBase2Unicode(asContent, giCodecUnicodeBase, 1);
}
