/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Düsterloh <mnenhy@tprac.de>.
 *
 * Portions created by the Initial Developer are Copyright (C) 2002-2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Karsten Düsterloh <mnenhy@tprac.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// +-
// |  These Mozilla MailNews functions are overloaded here:
// |    (none, but see mnenhy-headers-msgHdrViewOverlay-loader.js)
// |  Last checked for changes: 2004-11-24
// +-

// +-
// |  the contents of this file will be loaded into the Mnenhy.headers.* context!
// |  ie. foo() is available as goMnenhy.headers.foo()
// +-



// |
// |  constants
// |

// properties
const gkoPropertiesHeaders  = goMnenhy.InitStringBundle("chrome://mnenhy/locale/mnenhy-headers.properties");
const gksFailedMIDF         = goMnenhy.GetString("MIDF_failed",  gkoPropertiesHeaders);
const gksVersionMIDF        = goMnenhy.GetString("MIDF_version", gkoPropertiesHeaders);
const gksTitleMIDF          = goMnenhy.GetString("MIDF_title",   gkoPropertiesHeaders);
const gksFailedNG           = goMnenhy.GetString("NG_failed",    gkoPropertiesHeaders);

// expanded headers from RDF
const gksToggle             = "useToggle";
const gksKnown              = "header:known";
const gksFormat             = "format";
const gksIndex              = "index";
const gksDisplay            = "display";
const gksDefOut             = "text";

// expanded XUL elements
const gksFieldText          = "mail-headerfield";
const gksFieldMail          = "mail-emailheaderfield";
const gksFieldMulti         = "mail-multi-emailHeaderField";
// expanded XUL bindings
const gksClassRef           = "MnenhyRefField";
const gksClassNG            = "MnenhyNGField";
const gksClassText          = "MnenhyTextField";
const gksClassPersist       = "MnenhyToggleState";

// function names
const gksOutputFunction     = "outputFunction";
const gksfOutputRef         = "OutputReferences";
const gksfOutputNews        = "OutputNewsgroups";
const gksfOutputMail        = "OutputEmailAddresses";

// special newsgroup value
const gksPoster             = "poster";

// minimum usable version of messageidfinder
const gksMIDF               = "messageidfinder";
const gksMIDFMenu           = "messageidfinder-menu";
const gksMIDFVersionMinimum = "1.9";



// |
// |  variables
// |

// We need a global window context for most functions, because they
// will call other functions and use objects there. Since we may have
// several windows open, we need a pointer to the current one...
// (see respective calls to SetCurrentWindow)
var goWin = window;

// shortcut to Mnenhy;
var goMnenhy = null;

// robbed dtd strings
var goMnenhyDTDRobbing = null;

// if messageidfinder exists and has the right version, use it
// (will be inited in InitMsgHdrViewOverlay)
var goMIDF   = null; // the MIDF menu object


// |
// |  functions
// |

// set the current window context
function SetCurrentWindow(aoWin)
{
  this.goWin = aoWin || window;
  // set content window for SpaceHit, Mail Start Page etc.
  this.goWin.content = this.goWin.document.getElementById("messagepane").contentWindow;
}


// init MIDF etc.
function InitMsgHdrViewOverlay(aoMnenhy)
{
  this.goMnenhy = aoMnenhy;
  CheckMIDF();
}


// create an entry for asUniqueName in gExpandedHeaderList
// by reading respective data for asDisplayName from mnenhy.rdf
// if asDisplayName is not given, use asUniqueName instead
function CreateExpandedFromRDF(asUniqueName, asDisplayName)
{
  if (!asDisplayName)
    asDisplayName = asUniqueName;

  const kofOutput = {text: goWin.updateHeaderValue,
                     mail: goWin.OutputEmailAddresses,
                     news: OutputNewsgroups,
                     mid:  OutputReferences,
                     link: OutputLinkify };
  const kbToggle  = !!goMnenhy.m_oRDF.GetItemValue(gksToggle,  [gksKnown, asDisplayName]);
  const ksSort    =   goMnenhy.m_oRDF.GetItemValue(gksIndex,   [gksKnown, asDisplayName]) || asDisplayName;
  const ksOutput  =   goMnenhy.m_oRDF.GetItemValue(gksFormat,  [gksKnown, asDisplayName]);
  const kfOutput  = kofOutput[(ksOutput in kofOutput) ? ksOutput : gksDefOut];

  return {name:          asUniqueName,
          display:       asDisplayName,
          index:         ksSort,
          useToggle:     kbToggle,
          outputFunction:kfOutput};
}


function SortByIndexString(o1,o2)
{
  if (o1.index < o2.index)
    return -1;
  if (o1.index > o2.index)
    return  1;
  return 0;
}


function InitializeHeaderViewTables()
{
  var viewmode = GetHeaderView();
  if (viewmode == gkiViewNoHeaders)
  { // delete views
    goWin.gCollapsedHeaderView = {};
    goWin.gExpandedHeaderView  = {};
  }
  else
  { // Build HeaderViews

    // Remarks: * if (!gXXXHeaderView.length) is useless, because gXXXHeaderView is associative!
    //          * isEmailAddress is used nowhere and hence ignored here

    // CollapsedHeaderView
    // --> we stick to the list of headers predefined by Mozilla
    for (var index = 0; index < goWin.gCollapsedHeaderList.length; index++)
    {
      goWin.gCollapsedHeaderView[goWin.gCollapsedHeaderList[index].name] =
        new goWin.createHeaderEntry('collapsed', goWin.gCollapsedHeaderList[index]);
    }

    // ExpandedHeaderView
    // --> we prepare some headers in advance and sort them
    goWin.gExpandedHeaderList = []; // delete Mozilla's definitions

    var rsHeaderList = "";
    if (viewmode >= gkiViewCustomBase)
      rsHeaderList = goMnenhy.m_oRDF.GetNodeList("header:custom" + (viewmode - gkiViewCustomBase));
    else
    {
      rsHeaderList = ["subject", "from", "reply-to", "date", "to", "cc", "bcc", "newsgroups", "followup-to"];
      if (goWin.gShowOrganization)
        rsHeaderList.push("organization");
      if (goWin.gShowUserAgent)
        rsHeaderList.push("user-agent");
    }
    for (var iHeader in rsHeaderList)
      goWin.gExpandedHeaderList[iHeader] = CreateExpandedFromRDF(rsHeaderList[iHeader])
    goWin.gExpandedHeaderList.sort(SortByIndexString)
    // sync current gExpandedHeaderView and gExpandedHeaderList
    for (index = 0; index < goWin.gExpandedHeaderList.length; index++)
    {
      const ksHeaderName = goWin.gExpandedHeaderList[index].name;
      goWin.gExpandedHeaderView[ksHeaderName] = new CreateExpandedView(goWin.gExpandedHeaderList[index]);
    }
  } // Build HeaderViews
}


function OutputLinkify(aoHeaderEntry, aoHeaderValue)
{ // linkification is done via class, no special needs here
  goWin.updateHeaderValue(aoHeaderEntry, aoHeaderValue);
}


function OutputReferences(aoHeaderEntry, aoHeaderValue)
{ // do it like NS 4.x did: show linkified numbers instead of actual MID
  // <MID> => [1] -> news://<currentServer>/MID
  // this Link should be copyable and clicking should call the MIDF, if present
  aoHeaderEntry.enclosingBox.clearEmailAddresses();
  var rsRefs = aoHeaderValue.match(/[^<>\s]+/g);  // MIDs without <>
  for (var iRef = 0; iRef < rsRefs.length; ++iRef)
  {
    var oRef = {sMID:   rsRefs[iRef],
                iIndex: iRef + 1};
    aoHeaderEntry.enclosingBox.addAddressView(oRef);
  }
  aoHeaderEntry.enclosingBox.buildViews(-1);  // show all
}


function OutputNewsgroups(aoHeaderEntry, aoHeaderValue)
{ // do it like NS 4.x did: show linkified newsgroups
  // <NG> => [NG] -> news://<currentServer>/NG
  // this Link should be copyable and clicking should call the NG, if present
  aoHeaderEntry.enclosingBox.clearEmailAddresses()
  var rsNGs = aoHeaderValue.split(",");
  for (var iNG = 0; iNG < rsNGs.length; ++iNG)
  {
    aoHeaderEntry.enclosingBox.addAddressView(rsNGs[iNG]);
  }
  aoHeaderEntry.enclosingBox.buildViews(goWin.gNumAddressesToShow);  // usually show only 3
}


function UpdateMessageHeaders()
{ // reset x-face overrule of AOL buddy icons
  goMnenhy.Log("UpdateMessageHeaders")
  if ("gfSetFromBuddyIcon" in goWin)
  {
    goWin.setFromBuddyIcon = goWin.gfSetFromBuddyIcon;
    delete goWin.gfSetFromBuddyIcon;
  }

  // iterate over each header we received and see if we
  // have a matching entry in each header view table...
  for (var sHeaderName in goWin.currentHeaderData)
  {
    goMnenhy.Log("sHeaderName="+sHeaderName)
    var oCurrentHeader = goWin.currentHeaderData[sHeaderName];
    var oHeaderEntry   = null;
    // some headers need special attention
    switch (sHeaderName)
    {
      case "subject":
      { // show subject in window title bar
        try
        {
          if (goWin.gDBView.keyForFirstSelectedMessage == goWin.nsMsgKey_None)
          {
            var oFolder = null;
            if (goWin.gCurrentFolderUri)
              oFolder = goWin.RDF.GetResource(goWin.gCurrentFolderUri).QueryInterface(Components.interfaces.nsIMsgFolder);
            goWin.setTitleFromFolder(oFolder, oCurrentHeader.headerValue);
          }
        }
        catch(ignored){}
        break;
      }

      case "x-face":
      {
        if ("xface" in goMnenhy.headers)
        { // set x-face
          const ksFaceURL = goMnenhy.headers.xface.FaceURL(oCurrentHeader.headerValue);
          const koFromBuddyIcon = goWin.document.getElementById("fromBuddyIcon");
          koFromBuddyIcon.setAttribute("src", ksFaceURL);
          // overrule AOL buddy icons
          goWin.gfSetFromBuddyIcon = goWin.setFromBuddyIcon;
          goWin.setFromBuddyIcon = function(ignored){};
        }
      }
    }

    if (goWin.gCollapsedHeaderViewMode && !goWin.gBuiltCollapsedView)
    {
      if (sHeaderName in goWin.gCollapsedHeaderView)
      { // use existing view
        oHeaderEntry = goWin.gCollapsedHeaderView[sHeaderName];
      }
    }
    else if (!goWin.gCollapsedHeaderViewMode && !goWin.gBuiltExpandedView)
    {
      if (sHeaderName in goWin.gExpandedHeaderView)
      { // use existing view
        oHeaderEntry = goWin.gExpandedHeaderView[sHeaderName];
      }
      else
      { // non-precreated headers are only shown if the view mode is gkiViewAllHeaders
        // or if the 'real' name is valid for display
        const kiViewMode = GetHeaderView();
        if (    (kiViewMode == gkiViewAllHeaders)
             || (oCurrentHeader.headerName in goWin.gExpandedHeaderView))
        { // create new view
          const koExpandedListItem = CreateExpandedFromRDF(sHeaderName, oCurrentHeader.headerName);
          oHeaderEntry = new CreateExpandedView(koExpandedListItem)
          goWin.gExpandedHeaderView[sHeaderName] = oHeaderEntry;
        }
      } // create new view
    } // if we are in expanded view....

    if (oHeaderEntry)
    {
      oHeaderEntry.outputFunction(oHeaderEntry, oCurrentHeader.headerValue);
      oHeaderEntry.valid = true;
    }
  } // for (var sHeaderName in currentHeaderData)

  if (goWin.gCollapsedHeaderViewMode)
      goWin.gBuiltCollapsedView = true;
  else
      goWin.gBuiltExpandedView = true;

  // now update the view to make sure the right elements are visible
  goWin.updateHeaderViews();

  if ("FinishEmailProcessing" in goWin)
      goWin.FinishEmailProcessing();
}


// modify the messageHeaderSink's onStartHeaders function
// if we don't, the header view does not get updated when we change from all to custom
function OnStartHeaders()
{
  goMnenhy.Log("OnStartHeaders")

  if ("mSaveHdr" in goWin.messageHeaderSink)  // XXX TB 1.0 hack
  {
    goWin.messageHeaderSink.mSaveHdr = null;  // that's how it should be
    goWin.mSaveHdr = null;                    // that's what they did :(
  }

  var oElem = goWin.document.getElementById("msgHeaderView");
  if (GetHeaderView() == gkiViewNoHeaders)
  { // <sing>no headers, no box...</sing> ;-)
    oElem.setAttribute("collapsed","true");
  }
  else
  {
    oElem.removeAttribute("collapsed");

    // clear out any pending collected address timers...
    if (goWin.gCollectAddressTimer)
    {
      goWin.gCollectAddress = "";
      goWin.clearTimeout(goWin.gCollectAddressTimer);
      goWin.gCollectAddressTimer = null;
    }

    var iHeaderView = GetHeaderView();
    if (iHeaderView != goWin.giHeaderView)
    { // kill all headers to enforce rebuild
      goWin.hideHeaderView(goWin.gExpandedHeaderView);
      const koTopViewNode = goWin.document.getElementById("expandedHeaders");
      var oChild = koTopViewNode.lastChild, oNextChild;
      while (oChild)
      {
        oNextChild = oChild.previousSibling;
        if (oChild.id.search(/^Mnenhy/)) // do not kill our popups and HTML style source
          koTopViewNode.removeChild(oChild);
        oChild = oNextChild;
      }
      goWin.gExpandedHeaderView = {};
      goWin.initializeHeaderViewTables();
      goWin.giHeaderView = iHeaderView;
    }

    // set gViewAllHeaders (for backwards and Enigmail compatibility)
    goWin.gViewAllHeaders = true;

    goWin.ClearCurrentHeaders();
    goWin.gBuiltExpandedView = false;
    goWin.gBuiltCollapsedView = false;
    goWin.gBuildAttachmentsForCurrentMsg = false;
    goWin.gBuildAttachmentPopupForCurrentMsg = true;
    goWin.ClearAttachmentList();
    goWin.ClearEditMessageButton();

    // 2004-11-24 for Thunderbird 0.9
    if ("SetUpRemoteContentBar" in goWin)
      goWin.SetUpRemoteContentBar(null);
    // 2005-02-16 Bug #282429 for TB1.0+
    else if ("gMessageNotificationBar" in goWin && "clearMsgNotifications" in goWin.gMessageNotificationBar)
      goWin.gMessageNotificationBar.clearMsgNotifications();

    for (var index in goWin.gMessageListeners)
        goWin.gMessageListeners[index].onStartHeaders();
  }
}


// modify the messageHeaderSink's processHeaders function
function ProcessHeadersNew(arHeaderNames, arHeaderValues, abDontCollectAddress)
{ // Mozilla 1.8a and above
  goMnenhy.Log("ProcessHeadersNew")
  goWin.messageHeaderSink.onStartHeaders();
  // process each header
  while (arHeaderNames.hasMore())
  {
    var oHeader = {headerName:  arHeaderNames .getNext(),
                   headerValue: arHeaderValues.getNext()};
    if (oHeader.headerName)
      ProcessHeader(oHeader, abDontCollectAddress);
  }
  goWin.messageHeaderSink.onEndHeaders();
}
function ProcessHeadersOld(arHeaderNames, arHeaderValues, aiNumHeaders, abDontCollectAddress)
{ // Mozilla 1.7 and below
  goMnenhy.Log("ProcessHeadersOld")
  goWin.messageHeaderSink.onStartHeaders();
  // process each header
  var iIndex = 0;
  while (iIndex < aiNumHeaders)
  { // for consistancy sake, let's force all header names to be lower case so
    var oHeader = {headerName:  arHeaderNames [iIndex],
                   headerValue: arHeaderValues[iIndex]};
    ProcessHeader(oHeader, abDontCollectAddress);
    iIndex++;
  }
  goWin.messageHeaderSink.onEndHeaders();
}
// if there're multiple headers of the same kind, we usually show them all.
// but in case of to, cc and bcc, we merge
// !!! the bug fixed here is fixed in the Mozilla trunk since 2002-12-10 (->1.3a) !
function ProcessHeader(aoHeader, abDontCollectAddress)
{
  // for consistancy sake, let's force all header names to be lower case so
  // we don't have to worry about looking for: Cc and CC, etc.
  var sLowerCaseHeaderName = aoHeader.headerName.toLowerCase();
  aoHeader.headerName = sLowerCaseHeaderName;
  var sUniqueName = sLowerCaseHeaderName;
  var bOldHeader = (sUniqueName in goWin.currentHeaderData);
  if (bOldHeader)
  { // need to create an unique name
    // special handling for certain headers
    switch (sLowerCaseHeaderName)
    {
      case "to":
      case "cc":
      case "bcc":
          goWin.currentHeaderData[sLowerCaseHeaderName].headerValue = goWin.currentHeaderData[sLowerCaseHeaderName].headerValue
                                                                    + ',' + aoHeader.headerValue;
          break;

      default:
      {
        var iUID = 0;
        do
        {
          sUniqueName = sLowerCaseHeaderName + iUID;
          ++iUID;
        }
        while (sUniqueName in goWin.currentHeaderData);
        bOldHeader = false; // create this new header
      }
      break;
    }
  }

  if (!bOldHeader)
  {
    goWin.currentHeaderData[sUniqueName] = aoHeader;
  }

  if (sLowerCaseHeaderName == "from")
  {
    if (aoHeader.headerValue)
    {
      if (  (goWin.gCollectIncoming  && !abDontCollectAddress)
          ||(goWin.gCollectNewsgroup &&  abDontCollectAddress)
         )
      {
        if (!goWin.abAddressCollector) // global var
          goWin.abAddressCollector = Components.classes[goWin.abAddressCollectorContractID]
                                               .getService(Components.interfaces.nsIAbAddressCollecter);
        // collect, and add card if doesn't exist
        goWin.gCollectAddress = aoHeader.headerValue;
        goWin.gCollectAddressTimer = setTimeout(CallAddressCollector,
                                                2000,
                                                goWin.gCollectAddress,
                                                true);
      }
      else if (goWin.gCollectOutgoing)
      {
        if (!goWin.abAddressCollector) // global var
          goWin.abAddressCollector = Components.classes[goWin.abAddressCollectorContractID]
                                               .getService(Components.interfaces.nsIAbAddressCollecter);
        // collect, but only update existing cards
        goWin.gCollectAddress = aoHeader.headerValue;
        goWin.gCollectAddressTimer = setTimeout(CallAddressCollector,
                                                2000,
                                                goWin.gCollectAddress,
                                                false);
      }
    }
  }
}
function CallAddressCollector(asAddress, abCreateCard)
{ // make sanity checks because the window maybe closed before this is called
  if (("abAddressCollector" in goWin) && ("collectUnicodeAddress" in goWin.abAddressCollector))
  { // since about 1.8a, collectUnicodeAddress takes a third parameter, but superfluous
    // parameters do not matter in JS, so this should be backwards compatible
    goWin.abAddressCollector
         .collectUnicodeAddress(asAddress,
                                abCreateCard,
                                Components.interfaces.nsIAbPreferMailFormat.unknown);
  }
}


// init the global goMIDF object
function CheckMIDF()
{ // alas, the following fails sometimes... :|
  var sMIDFVersionCurrent = InstallTrigger.getVersion(gksMIDF);
  var bMIDF = !!sMIDFVersionCurrent;
  if (!bMIDF)
  { // alas! the current version can be null even though MIDF is installed
    // (mozver.dat does not seem to be very reliable. :( )
    // So let's see if there's a MIDF version installed that already has the
    // version in its chrome.rdf entry...

    // first check the profile
    var oRDF = goMnenhy.uninstall.GetChromeRDF(true);
    if (oRDF)
      sMIDFVersionCurrent = oRDF.GetItemValue("version", ["package", gksMIDF], true);
    if (!sMIDFVersionCurrent)
    { // not found, check the app chrome dir
      oRDF = goMnenhy.uninstall.GetChromeRDF(false);
      if (oRDF)
        sMIDFVersionCurrent = oRDF.GetItemValue("version", ["package", gksMIDF], true);
    }
    bMIDF = !!sMIDFVersionCurrent;
  }

  if (bMIDF)
  { // a MIDF exists, so check version
    bMIDF = (sMIDFVersionCurrent >= gksMIDFVersionMinimum);

    if (bMIDF)
      goMIDF = goWin.document.getElementById(gksMIDFMenu);
  }
}




function SearchContextForMID(oNode)
{
  return (oNode.id == gksMIDFMenu) ? NodeFilter.FILTER_ACCEPT : NodeFilter.FILTER_REJECT;
}
function OnPopupShowingRef(event)
{
  if (!goMIDF)
    CheckMIDF();

  if (goMIDF)
  {
    var oMenuPopUp = event.originalTarget;
    if (oMenuPopUp.id != "messageidfinder-popup")
    {
      var oTreeWalker = goWin.document.createTreeWalker(oMenuPopUp,
                                                        NodeFilter.SHOW_ELEMENT,
                                                        SearchContextForMID,
                                                        false);
      var bNoMIDF = true;
      // actually, nextNode() should just return null if there's no such node...
      try{bNoMIDF = !oTreeWalker.nextNode();}catch(ignored){}
      if (bNoMIDF)
      {
        // +-
        // |  call MIDF
        goWin.AddNewsServers();
        // +-
        var oMIDFinderMenu = goMIDF.cloneNode(true);
        oMenuPopUp.appendChild(goWin.document.createElement("menuseparator"));
        oMenuPopUp.appendChild(oMIDFinderMenu);
      } // add MIDF menuitem
    } // no MIDF menuitem
  } // have MIDF
}


function OnPopupShowingText(event)
{
  var oMenuPopUp = event.originalTarget;
  if ((oMenuPopUp.id == "MnenhyTextPopup") && (!("mnenhyCreatedText" in oMenuPopUp)) && goWin)
  {
    var oEnc = goWin.document.getElementById("mnenhyCodecsEncMenu");
    var oDec = goWin.document.getElementById("mnenhyCodecsDecMenu");
    if (oEnc && oDec)
    {
      oMenuPopUp.appendChild(goWin.document.createElement("menuseparator"));
      oMenuPopUp.appendChild(oEnc.cloneNode(true));
      oMenuPopUp.appendChild(oDec.cloneNode(true));
      oMenuPopUp.mnenhyCreatedText = true;
    }
  }
}


// prefix --> the name of the view (i.e. "collapsed", "expanded")
// headerListInfo --> entry from a header list.
function CreateExpandedView(oHeaderEntry)
{
  if (!oHeaderEntry)
    return;

  var sHeaderName    = oHeaderEntry.name;
  var sPartialIDName = "expanded" + sHeaderName;
  this.enclosingBox  = goWin.document.getElementById(sPartialIDName + "Box");
  this.textNode      = goWin.document.getElementById(sPartialIDName + "Value");
  this.isValid       = false;

  // if we already have a view box, add our class,
  // else create a new box
  if (this.enclosingBox)
  { // class override
    AddExpandedBoxFeatures(this.enclosingBox, oHeaderEntry)
  }
  else
  { // new box
    this.enclosingBox = new CreateExpandedBox(oHeaderEntry);
  }

  if ("useToggle" in oHeaderEntry)
  {
    this.useToggle = oHeaderEntry.useToggle;
    // find the toggle icon in the document
    if (this.useToggle)
    {
      this.textNode = this.enclosingBox.emailAddresses;
    }
  }
  else
   this.useToggle = false;

  if ("useShortView" in oHeaderEntry)
  {
    this.useShortView = oHeaderEntry.useShortView;
    this.enclosingBox.emailAddressNode = this.textNode;
  }
  else
    this.useShortView = false;

  if (gksOutputFunction in oHeaderEntry)
    this.outputFunction = oHeaderEntry.outputFunction;
  else
    this.outputFunction = updateHeaderValue;
}


function CreateExpandedBox(oHeaderEntry)
{
  if (!oHeaderEntry)
    return null;

  // create element according to outputFunction
  var sOutputFunction = "";   // use default output
  if ((gksOutputFunction in oHeaderEntry) && (oHeaderEntry.outputFunction))
    sOutputFunction = oHeaderEntry.outputFunction.name;

  var sNewElement;
  switch (sOutputFunction)
  {
    case gksfOutputRef:
      sNewElement = gksFieldMulti;
      break;

    case gksfOutputNews:
      sNewElement = gksFieldMulti;
      break;

    case gksfOutputMail:
      if ((!("useToggle" in oHeaderEntry)) || (!oHeaderEntry.useToggle))
        sNewElement = gksFieldMail;  // one address
      else
        sNewElement = gksFieldMulti; // multiple
      break;

    default:
      sNewElement = gksFieldText;
  }

  // create new view and set attributes
  // 1. check mnenhy.rdf for user-defined header display names
  var sDisplayName = goMnenhy.m_oRDF.GetItemValue(gksDisplay, [gksKnown, oHeaderEntry.display]);
  if (sDisplayName)
    sDisplayName += ": ";
  // 2. else look for Mozilla-defined localized header names
  else
  {
    if (!goMnenhyDTDRobbing)
      goMnenhyDTDRobbing = goWin.document.getElementById("mnenhydtdrobbing");
    if (goMnenhyDTDRobbing)
    {
      const ksHeaderAttr = oHeaderEntry.name.replace("-", "");
      sDisplayName = goMnenhyDTDRobbing.getAttribute(ksHeaderAttr);
    }
  }
  // 3. else create our own nice display
  if (!sDisplayName)
    sDisplayName = goMnenhy.Capitalize(oHeaderEntry.display)+ ": ";

  const ksBoxIdName   = "expanded" + oHeaderEntry.name + "Box";
  var oNewView = goWin.document.createElement(sNewElement);
  oNewView.setAttribute("id",             ksBoxIdName);
  oNewView.setAttribute("label",          sDisplayName);
  oNewView.setAttribute("keywordrelated", "true");
  oNewView.collapsed = true;

  AddExpandedBoxFeatures(oNewView, oHeaderEntry);

  // insert this new element at the correct position into the view...
  var oTopViewNode = goWin.document.getElementById("expandedHeaders");
  oTopViewNode.appendChild(oNewView);
  return oNewView;
}


// set new features like class, context, etc. according to the output function
function AddExpandedBoxFeatures(oBox, oHeaderEntry)
{
  if (oBox && oHeaderEntry)
  {
    var sOutputFunction = "";   // use default output
    if ((gksOutputFunction in oHeaderEntry) && (oHeaderEntry.outputFunction))
      sOutputFunction = oHeaderEntry.outputFunction.name;

    // add new class
    var sFunctionClass, bMIDFhandler;
    switch (sOutputFunction)
    {
      case gksfOutputRef:
        sFunctionClass = gksClassRef;
        bMIDFhandler   = false;
        break;
      case gksfOutputNews:
        sFunctionClass = gksClassNG;
        bMIDFhandler   = false;
        break;
      case gksfOutputMail:
        sFunctionClass = gksClassPersist;
        bMIDFhandler   = false;
        break;
      default:
        sFunctionClass = gksClassText;
        bMIDFhandler   = !!goMIDF;
    }

    // add class
    var sNewClass = oBox.getAttribute("class")
    if (sFunctionClass && (String(sNewClass).indexOf(sFunctionClass) < 0))
    {
      sNewClass += " " + sFunctionClass;
      sNewClass = sNewClass.replace(/^\s+|\s+$/g, "");      // trim whitespace
      oBox.setAttribute("class", sNewClass);
    }

    // after work party ;-)
    if (sFunctionClass == gksClassText)
    { // now that we ain't partying in the switch default branch anymore,
      // keep the following code commented out for future reference only
      //      // Since we're constructing the oBox before it is shown, the XBL properties
      //      // aren't in place yet and assignment to them creates object members!
      //      // You can't |delete| properties, so it's safe to |delete| this members...
      //      // (I *really* do *hate* such bugs! :( )
      //      delete oBox.doLinkify;
      oBox.doLinkify = (sOutputFunction   == "OutputLinkify");
      // to allow for custom header styles, add headername as attribute
      // e.g.
      //      .MnenhyTextFieldHTML[header="subject"]
      //      {
      //        color:yellow!important;
      //      }
      oBox.setAttribute("header", oHeaderEntry.name);
    }

    // add MIDF, if reasonable and possible
    if (bMIDFhandler)
      oBox.addEventListener("popupshowing", OnPopupShowingRef, false);
  }
}


function OnMnenhyRefClick(event)
{
  if (!event.button)
  { // left click
    if (!goMIDF)
      CheckMIDF();

    if (goMIDF)
    { // correct MIDF version installed
//    goMnenhy.SetWaitCursor(goWin);  // this crashes in recent builds?!?
      // +-
      // |  call MIDF
      goWin.AddNewsServers();
      var oMsgHdr = goWin.OpenMessageByMessageID(event.originalTarget.getAttribute("sMID"));
      // +-
//    goMnenhy.UnsetWaitCursor(goWin);

      if (!oMsgHdr)
        goMnenhy.AlertService(gksFailedMIDF, null, goWin);
    }
    else
    {
      goMnenhy.Alert(gksVersionMIDF, gksTitleMIDF, goWin)
    }
  }
}


function OnMnenhyNGClick(event)
{
  if (!event.button)
  {
    const ksNG = event.originalTarget.getAttribute("sNG");
    if (ksNG == gksPoster)
    { // if clicked NG is poster, reply to sender
      goWin.goDoCommand('cmd_replySender');
    }
    else
    {
      const koServer = goWin.GetLoadedMsgFolder().server;
      const ksURI    = koServer.serverURI + "/" + ksNG;

      goMnenhy.SetWaitCursor(goWin);
      try
      {
        goWin.SelectFolder(ksURI);
      }
      catch(ex)
      { // could not select folder, maybe 'cos it ain't there yet
        goMnenhy.UnsetWaitCursor(goWin);
        if (ksURI.substr(0,7) == "news://")
        { // only try if in news folder
          const koSubServ = koServer.QueryInterface(Components.interfaces.nsISubscribableServer);
          if (GetSubscribeConfirmation(ksNG))
          {
            goMnenhy.SetWaitCursor(goWin);
            try
            {
              koSubServ.subscribe(ksNG);
              koSubServ.commitSubscribeChanges();
              goWin.SelectFolder(ksURI);  // retry selecting folder
            }
            catch(ex)
            {
              goMnenhy.UnsetWaitCursor(goWin);
              const ksFailed = gksFailedNG.replace("%ng", ksNG);
              goMnenhy.AlertService(ksFailed, null, goWin);
            } // no group found on server
          } // subscribing allowed
        } // we are on a news server
      }
      goMnenhy.UnsetWaitCursor(goWin);
    }
  }
}


function GetSubscribeConfirmation(sNG)
{

  try
  {
    const koNewsBundle = goMnenhy.InitStringBundle("chrome://messenger/locale/news.properties");
    var s = koNewsBundle.GetStringFromName("autoSubscribeText").replace("%1$S", sNG);
    return goMnenhy.ConfirmService(s, null, goWin);
  }
  catch(error)
  {
    return false;
  }
}


function PrettifyMID(sMID)
{
  return "<" + sMID + ">";
}


// CopyMID to the clipboard
function CopyMID(aoMID, abAll)
{
  if (aoMID)
  {
    var sMID = "";
    if (abAll)
    {
      var roMIDs = aoMID.parentNode.parentNode.mAddresses;
      for (var i = 0; i < roMIDs.length; ++i)
        sMID += PrettifyMID(roMIDs[i].sMID) + " ";
      sMID = sMID.replace(/\s+$/, "");
    }
    else
    {
      sMID = PrettifyMID(aoMID.getAttribute("sMID"));
    }
    goMnenhy.CopyStringToClipboard(sMID);
  }
}


// CopyNG to the clipboard
function CopyNG(aoNG, abAll)
{
  if (aoNG)
  {
    var sNG = "";
    if (abAll)
    {
      var roNGs = aoNG.parentNode.parentNode.mAddresses;
      for (var i = 0; i < roNGs.length; ++i)
        sNG += roNGs[i] + ",";
      sNG = sNG.replace(/,+$/, "");
    }
    else
    {
      sNG = aoNG.getAttribute("sNG");
    }
    goMnenhy.CopyStringToClipboard(sNG);
  }
}


// CopyText to the clipboard
function CopyText(aoText, abAll)
{
  if (aoText)
  {
    var sText = "";
    if (!abAll)
    { // if selection requested and non-empty, copy it (text only!)
      const oSel = aoText.ownerDocument.defaultView.getSelection();
      if (oSel)
        sText = oSel.toString();
    }
    if (!sText)
    { // if no selection requested or empty, copy original header value
      const ksObserverID = aoText.ownerDocument.documentElement.getAttribute("mnenhyObserverID");
      sText = goMnenhy.GetObserver(ksObserverID).headerValue;
    }
    goMnenhy.CopyStringToClipboard(sText);
  }
}