/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Düsterloh <mnenhy@tprac.de>.
 *
 * Portions created by the Initial Developer are Copyright (C) 2002-2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Karsten Düsterloh <mnenhy@tprac.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function InitMnenhyPrefMain(aoPage)
{ 
  parent.gksMnenhyMainPanelURL = "chrome://mnenhy/content/mnenhy-pref-main.xul";
  // init this panel
  parent.initPanel(parent.gksMnenhyMainPanelURL);
  // register pref exit function
  parent.hPrefWindow.registerOKCallbackFunc(ExitMnenhyPrefMain);
  // set the Mnenhy version as header title description
  const koHeader = top.document.getElementById("header")
  if (koHeader)
  {
    const ksVersion = "v" + goMnenhy.ksMnenhyVersion;
    koHeader.setAttribute("description", ksVersion);
  }
}


function UnloadMnenhyPrefMain()
{ // remove version from header
  if (top.document) // the opening document may have been gone...
  {
    const koHeader = top.document.getElementById("header")
    if (koHeader)
      koHeader.removeAttribute("description");
  }
}


function SetFields(aoPageData)
{
  const koDM = parent.hPrefWindow.wsm.dataManager;
  var oData = koDM.getPageData(parent.gksMnenhyMainPanelURL);
  if ("initialized" in oData && oData.initialized)
    return true;   // use stored data

  if ("elementIDs" in oData)
    oData = oData.elementIDs; // bird widget state manager
  else
    oData = aoPageData;       // Seamonkey widget state manager

  // init the page data
  const krPackages = goMnenhy.installed.GetAllPackages(true);
  var rPack, sPack, oCheck;
  for (var i in krPackages)
  {
    rPack  = krPackages[i];
    sPack  = rPack[goMnenhy.installed.kiIndexInfoName];
    oCheck = document.getElementById(sPack);
    if (oCheck)
    { // The Mnenhy base package is always installed
      // and can never be disabled here (the ID 'mnenhy' must not exist).
      // A package state can only be edited if the package is allowed and installed.
      oData[sPack] = {disabled:    !rPack[goMnenhy.installed.kiIndexInfoAllowed]
                                || !rPack[goMnenhy.installed.kiIndexInfoInstall], 
                      checked:      rPack[goMnenhy.installed.kiIndexInfoActive ] };
    }
  }
  return true;
}


function ExitMnenhyPrefMain()
{ // Change 'disabled' state according to checkboxes.
  // Since there is no function in nsIXULChromeRegistry to read the 'disabled'
  // state, we always just write everything that might have changed ;-)
  const koDM = parent.hPrefWindow.wsm.dataManager;
  var oData = koDM.getPageData(parent.gksMnenhyMainPanelURL);
  if ("initialized" in oData && oData.initialized)
  {
    if ("elementIDs" in oData)
      oData = oData.elementIDs; // bird widget state manager

    var oPack, bInstalled, bActive;
    for (var sPack in oData)
    {
      if (sPack == "initialized")
        continue;

      oPack = oData[sPack];
      if (!oPack.disabled)
        parent.hPrefWindow.chromeRegistry.setAllowOverlaysForPackage(sPack, oPack.checked);
    }
  } // is inited
}


function UninstallMnenhy()
{
  // To remove Mnenhy completely we would have to
  //  1. remove the entries in $MozillaProgramDir/chrome/installed-chrome.txt
  //  2. remove the entries in both /chrome/chrome.rdf files
  //  3. remove the entries in both /chrome/overlayinfo/ directories
  //  4. remove the mnenhy*.jar files
  //  5. remove mnenhy*.rdf in the profile
  //  6. remove the prefs.js hierarchy extensions.mnenhy.*
  // We will skip 5. and 6.
  // But we will mark Mnenhy for deinstallation in the Aviary's Extension Manager
  if (goMnenhy.ConfirmService(goMnenhy.GetString("MnDialog_UninstMnenhy"), null, this))
  {
    goMnenhy.SetWaitCursor(this);
    // unload mnenhy.rdf, so that resetting the view mode will not get saved there
    goMnenhy.m_oRDF = null;
    // set view mode back to normal (if installed)
    if ("headers" in goMnenhy && "SetHeaderView" in goMnenhy.headers)
      goMnenhy.headers.SetHeaderView(goMnenhy.headers.gkiViewNormalHeaders);

    // reset the User-Agent
    if (!goMnenhy.GetPref("extensions.mnenhy.useragent.nooverride", goMnenhy.kiPrefBool))
    { // we're allowed to touch the User-Agent
      var sUA = navigator.userAgent;
      if (sUA.search(goMnenhy.kxMnenhyUA) >= 0)
      { // remove Mnenhy
        sUA = sUA.replace(goMnenhy.kxMnenhyUA, "");
        // we can remove the entire pref if it was not faked
        if (!goMnenhy.GetPref("extensions.mnenhy.useragent.fake", goMnenhy.kiPrefBool))
          sUA = null;      
        goMnenhy.SetPref("general.useragent.override", goMnenhy.kiPrefString, sUA);
        goMnenhy.SetPref("extensions.mnenhy.useragent.fake", goMnenhy.kiPrefBool, null);      
      }
    }

    // remove Mnenhy from the Aviary's Extension Manager
    try
    {
      const ksEMClassID = "@mozilla.org/extensions/manager;1";
      const ksEMIID     = "nsIExtensionManager";
      if (ksEMClassID in Components.classes && ksEMIID in Components.interfaces)
      {
        Components.classes[ksEMClassID]
                  .getService(Components.interfaces[ksEMIID])
                  .uninstallExtension(goMnenhy.ksMnenhyUUID);
      }
      goMnenhy.Log("UninstallMnenhy: successfully removed from EM");
    }
    catch(oErr)
    {
      goMnenhy.Log("UninstallMnenhy: removing from EM failed:" + oErr);
    }
    // uninstall Mnenhy and shutdown Mozilla
    goMnenhy.Log("Deinstalling Mnenhy")
    goMnenhy.uninstall.Uninstall(goMnenhy.installed.GetAllPackages());
    //goMnenhy.UnsetWaitCursor(this);
  }
}


